@@ -29,7 +29,8 @@ from ._cache import DNSCache, _UniqueRecordsType
 from ._dns import DNSAddress, DNSNsec, DNSPointer, DNSQuestion, DNSRRSet, DNSRecord
 from ._history import QuestionHistory
 from ._logger import log
-from ._protocol import DNSIncoming, DNSOutgoing
+from ._protocol.incoming import DNSIncoming
+from ._protocol.outgoing import DNSOutgoing
 from ._services.info import ServiceInfo
 from ._services.registry import ServiceRegistry
 from ._updates import RecordUpdate, RecordUpdateListener
@@ -226,6 +227,19 @@ class _QueryResponse:
         return bool(maybe_entry and self._now - maybe_entry.created < _ONE_SECOND)
 
 
+def _get_address_and_nsec_records(service: ServiceInfo, now: float) -> Set[DNSRecord]:
+    """Build a set of address records and NSEC records for non-present record types."""
+    seen_types: Set[int] = set()
+    records: Set[DNSRecord] = set()
+    for dns_address in service.dns_addresses(created=now):
+        seen_types.add(dns_address.type)
+        records.add(dns_address)
+    missing_types: Set[int] = _ADDRESS_RECORD_TYPES - seen_types
+    if missing_types:
+        records.add(construct_nsec_record(service.server, list(missing_types), now))
+    return records
+
+
 class QueryHandler:
     """Query the ServiceRegistry."""
 
@@ -260,21 +274,9 @@ class QueryHandler:
             if known_answers.suppresses(dns_pointer):
                 continue
             additionals: Set[DNSRecord] = {service.dns_service(created=now), service.dns_text(created=now)}
-            additionals |= self._get_address_and_nsec_records(service, now)
+            additionals |= _get_address_and_nsec_records(service, now)
             answer_set[dns_pointer] = additionals
 
-    def _get_address_and_nsec_records(self, service: ServiceInfo, now: float) -> Set[DNSRecord]:
-        """Build a set of address records and NSEC records for non-present record types."""
-        seen_types: Set[int] = set()
-        records: Set[DNSRecord] = set()
-        for dns_address in service.dns_addresses(created=now):
-            seen_types.add(dns_address.type)
-            records.add(dns_address)
-        missing_types: Set[int] = _ADDRESS_RECORD_TYPES - seen_types
-        if missing_types:
-            records.add(construct_nsec_record(service.server, list(missing_types), now))
-        return records
-
     def _add_address_answers(
         self,
         name: str,
@@ -324,14 +326,14 @@ class QueryHandler:
             self._add_address_answers(question.name, answer_set, known_answers, now, type_)
 
         if type_ in (_TYPE_SRV, _TYPE_TXT, _TYPE_ANY):
-            service = self.registry.async_get_info_name(question.name)  # type: ignore
+            service = self.registry.async_get_info_name(question.name)
             if service is not None:
                 if type_ in (_TYPE_SRV, _TYPE_ANY):
                     # Add recommended additional answers according to
                     # https://tools.ietf.org/html/rfc6763#section-12.2.
                     dns_service = service.dns_service(created=now)
                     if not known_answers.suppresses(dns_service):
-                        answer_set[dns_service] = self._get_address_and_nsec_records(service, now)
+                        answer_set[dns_service] = _get_address_and_nsec_records(service, now)
                 if type_ in (_TYPE_TXT, _TYPE_ANY):
                     dns_text = service.dns_text(created=now)
                     if not known_answers.suppresses(dns_text):
@@ -389,7 +391,7 @@ class RecordManager:
         for listener in self.listeners:
             listener.async_update_records(self.zc, now, records)
 
-    def async_updates_complete(self) -> None:
+    def async_updates_complete(self, notify: bool) -> None:
         """Used to notify listeners of new information that has updated
         a record.
 
@@ -399,7 +401,8 @@ class RecordManager:
         """
         for listener in self.listeners:
             listener.async_update_records_complete()
-        self.zc.async_notify_all()
+        if notify:
+            self.zc.async_notify_all()
 
     def async_updates_from_response(self, msg: DNSIncoming) -> None:
         """Deal with incoming response packets.  All answers
@@ -457,15 +460,16 @@ class RecordManager:
         # zc.get_service_info will see the cached value
         # but ONLY after all the record updates have been
         # processsed.
+        new = False
         if other_adds or address_adds:
-            self.cache.async_add_records(itertools.chain(address_adds, other_adds))
+            new = self.cache.async_add_records(itertools.chain(address_adds, other_adds))
         # Removes are processed last since
         # ServiceInfo could generate an un-needed query
         # because the data was not yet populated.
         if removes:
             self.cache.async_remove_records(removes)
         if updates:
-            self.async_updates_complete()
+            self.async_updates_complete(new)
 
     def _async_mark_unique_cached_records_older_than_1s_to_expire(
         self, unique_types: Set[Tuple[str, int, int]], answers: Iterable[DNSRecord], now: float
@@ -490,6 +494,12 @@ class RecordManager:
 
         This function is not threadsafe and must be called in the eventloop.
         """
+        if not isinstance(listener, RecordUpdateListener):
+            log.error(
+                "listeners passed to async_add_listener must inherit from RecordUpdateListener;"
+                " In the future this will fail"
+            )
+
         self.listeners.append(listener)
 
         if question is None:
@@ -507,12 +517,12 @@ class RecordManager:
         This function must be run from the event loop.
         """
         now = current_time_millis()
-        records: List[RecordUpdate] = []
-        for question in questions:
-            for record in self.cache.async_entries_with_name(question.name):
-                if not record.is_expired(now) and question.answered_by(record):
-                    records.append(RecordUpdate(record, None))
-
+        records: List[RecordUpdate] = [
+            RecordUpdate(record, None)
+            for question in questions
+            for record in self.cache.async_entries_with_name(question.name)
+            if not record.is_expired(now) and question.answered_by(record)
+        ]
         if not records:
             return
         listener.async_update_records(self.zc, now, records)
