@@ -23,7 +23,7 @@
 import ipaddress
 import random
 import socket
-from typing import Any, Dict, List, Optional, TYPE_CHECKING, Union, cast
+from typing import Any, Dict, List, Optional, Set, TYPE_CHECKING, Union, cast
 
 from .._dns import DNSAddress, DNSPointer, DNSQuestionType, DNSRecord, DNSService, DNSText
 from .._exceptions import BadTypeInNameException
@@ -99,7 +99,23 @@ class ServiceInfo(RecordUpdateListener):
       where the peer is connected to
     """
 
-    text = b''
+    __slots__ = (
+        "text",
+        "type",
+        "_name",
+        "key",
+        "_ipv4_addresses",
+        "_ipv6_addresses",
+        "port",
+        "weight",
+        "priority",
+        "server",
+        "server_key",
+        "_properties",
+        "host_ttl",
+        "other_ttl",
+        "interface_index",
+    )
 
     def __init__(
         self,
@@ -122,6 +138,7 @@ class ServiceInfo(RecordUpdateListener):
             raise TypeError("addresses and parsed_addresses cannot be provided together")
         if not type_.endswith(service_type_name(name, strict=False)):
             raise BadTypeInNameException
+        self.text = b''
         self.type = type_
         self._name = name
         self.key = name.lower()
@@ -308,20 +325,18 @@ class ServiceInfo(RecordUpdateListener):
 
     def _process_records_threadsafe(self, zc: 'Zeroconf', now: float, records: List[RecordUpdate]) -> None:
         """Thread safe record updating."""
-        update_addresses = False
+        seen_addresses: Set[bytes] = set()
         for record_update in records:
-            if isinstance(record_update[0], DNSService):
-                update_addresses = True
-            self._process_record_threadsafe(record_update[0], now)
-
-        # Only update addresses if the DNSService (.server) has changed
-        if not update_addresses:
-            return
-
-        for record in self._get_address_records_from_cache(zc):
+            record = record_update.new
+            if isinstance(record, DNSAddress):
+                seen_addresses.add(record.address)
             self._process_record_threadsafe(record, now)
+        for record in self._get_address_records_from_cache(zc):
+            if record.address not in seen_addresses:
+                self._process_record_threadsafe(record, now)
 
     def _process_record_threadsafe(self, record: DNSRecord, now: float) -> None:
+        """Thread safe record updating."""
         if record.is_expired(now):
             return
 
@@ -343,6 +358,11 @@ class ServiceInfo(RecordUpdateListener):
                     self.interface_index = record.scope_id
             return
 
+        if isinstance(record, DNSText):
+            if record.key == self.key:
+                self._set_text(record.text)
+            return
+
         if isinstance(record, DNSService):
             if record.key != self.key:
                 return
@@ -352,11 +372,6 @@ class ServiceInfo(RecordUpdateListener):
             self.port = record.port
             self.weight = record.weight
             self.priority = record.priority
-            return
-
-        if isinstance(record, DNSText):
-            if record.key == self.key:
-                self._set_text(record.text)
 
     def dns_addresses(
         self,
@@ -413,12 +428,15 @@ class ServiceInfo(RecordUpdateListener):
             created,
         )
 
-    def _get_address_records_from_cache(self, zc: 'Zeroconf') -> List[DNSRecord]:
+    def _get_address_records_from_cache(self, zc: 'Zeroconf') -> List[DNSAddress]:
         """Get the address records from the cache."""
-        return [
-            *zc.cache.get_all_by_details(self.server, _TYPE_A, _CLASS_IN),
-            *zc.cache.get_all_by_details(self.server, _TYPE_AAAA, _CLASS_IN),
-        ]
+        return cast(
+            "List[DNSAddress]",
+            [
+                *zc.cache.get_all_by_details(self.server, _TYPE_A, _CLASS_IN),
+                *zc.cache.get_all_by_details(self.server, _TYPE_AAAA, _CLASS_IN),
+            ],
+        )
 
     def load_from_cache(self, zc: 'Zeroconf') -> bool:
         """Populate the service info from the cache.
@@ -467,6 +485,8 @@ class ServiceInfo(RecordUpdateListener):
         """Returns true if the service could be discovered on the
         network, and updates this object with details discovered.
         """
+        if not zc.started:
+            await zc.async_wait_for_start()
         if self.load_from_cache(zc):
             return True
 
@@ -475,7 +495,6 @@ class ServiceInfo(RecordUpdateListener):
         delay = _LISTENER_TIME
         next_ = now
         last = now + timeout
-        await zc.async_wait_for_start()
         try:
             zc.async_add_listener(self, None)
             while not self._is_complete:
